Attribute VB_Name = "Intersection_boite_3D"
Option Explicit

'---------------------------------------------------------------------------
' Point en 3D
'---------------------------------------------------------------------------
Public Type Point3
    X As Double
    Y As Double
    Z As Double
End Type

'---------------------------------------------------------------------------
'Boite en 3D
'---------------------------------------------------------------------------
' Cette notation Centre et axis est plus simple que de donner les sommets de
' la boite . Idem pour les longueurs ca permet de faire des tests sans recalculer des
' vecteurs

Public Type Box3
   Centre As Point3
   Axes(2) As Point3
   Longueurs(2) As Double
End Type

' transforme un vecteur en vecteur unitaire
Function VecteurUnitaire(P1 As Point3) As Point3
Dim Norm As Double
Norm = Sqr((P1.X) ^ 2 + (P1.Y) ^ 2 + (P1.Z) ^ 2)
If Norm = 0 Then
    Exit Function
End If
    
    VecteurUnitaire.X = P1.X / Norm
    VecteurUnitaire.Y = P1.Y / Norm
    VecteurUnitaire.Z = P1.Z / Norm

End Function

'Produit vectoriel
Function VecProd(P1 As Point3, P2 As Point3) As Point3
Dim P4 As Point3

 P4.X = (P1.Y * P2.Z) - (P1.Z * P2.Y)
 P4.Y = (P1.Z * P2.X) - (P1.X * P2.Z)
 P4.Z = (P1.X * P2.Y) - (P1.Y * P2.X)
 VecProd = P4
 
End Function

'Produit Scalaire
Function Dot(p As Point3, q As Point3) As Double
    Dot = p.X * q.X + p.Y * q.Y + p.Z * q.Z
End Function

' Soustraction de vecteur
Function VecSub(P1 As Point3, P2 As Point3) As Point3
 VecSub.X = P1.X - P2.X
 VecSub.Y = P1.Y - P2.Y
 VecSub.Z = P1.Z - P2.Z
End Function

'----------------------------------------------------------------------------
Function TestIntersectionBoite(Box0 As Box3, Box1 As Box3) As Boolean
    Const fCutoff As Double = 0.9999999  'f '
    Dim ExistePaireParallele As Boolean
    Dim i As Integer
    Dim akA(2) As Point3
    Dim akB(2) As Point3
    Dim afEA(2) As Double
    Dim afEB(2) As Double
    Dim kD As Point3
    
    Dim aafC(2, 2) As Double    ' matrice C = A^T B, c_{ij} = Dot(A_i,B_j)
    Dim aafAbsC(2, 2) As Double   ' |C(i,j)|
    Dim afAD(2) As Double         ' Dot(A_i,D)
    Dim fR0 As Double  ' Rayon interval et distance entre centres
    Dim fR1 As Double  ' Rayon interval et distance entre centres
    Dim fR As Double  ' Rayon interval et distance entre centres
    Dim fR01 As Double ' = R0 + R1
    
    
    ' Init variables de calcul
    For i = 0 To 2
        ' transforme en vecteur unitaires
        akA(i) = VecteurUnitaire(Box0.Axes(i))
        akB(i) = VecteurUnitaire(Box1.Axes(i))
        afEA(i) = Box0.Longueurs(i)
        afEB(i) = Box1.Longueurs(i)
    Next i

    ' Calcul le vecteurs entre les deux centres des boites D = C1-C0
    kD = VecSub(Box1.Centre, Box0.Centre)

    ' Axes C0+t*A0
    For i = 0 To 2
        aafC(0, i) = Dot(akA(0), akB(i)) '
        aafAbsC(0, i) = Abs(aafC(0, i)) '
        If (aafAbsC(0, i) > fCutoff) Then
            ExistePaireParallele = True
        End If
    Next i
    
    afAD(0) = Dot(akA(0), kD) '
    fR = Abs(afAD(0)) '
    fR1 = afEB(0) * aafAbsC(0, 0) + afEB(1) * aafAbsC(0, 1) + afEB(2) * aafAbsC(0, 2) '
    fR01 = afEA(0) + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False
        Exit Function
    End If

    ' Axes C0+t*A1
    For i = 0 To 2
        aafC(1, i) = Dot(akA(1), akB(i)) '
        aafAbsC(1, i) = Abs(aafC(1, i)) '
        If (aafAbsC(1, i) > fCutoff) Then
            ExistePaireParallele = True '
        End If
    Next i
    
    afAD(1) = Dot(akA(1), kD) '
    fR = Abs(afAD(1)) '
    fR1 = afEB(0) * aafAbsC(1, 0) + afEB(1) * aafAbsC(1, 1) + afEB(2) * aafAbsC(1, 2) '
    fR01 = afEA(1) + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If

    ' Axes C0+t*A2
    For i = 0 To 2
        aafC(2, i) = Dot(akA(2), akB(i)) '
        aafAbsC(2, i) = Abs(aafC(2, i)) '
        If (aafAbsC(2, i) > fCutoff) Then
            ExistePaireParallele = True '
        End If
    Next i
    
    afAD(2) = Dot(akA(2), kD) '
    fR = Abs(afAD(2)) '
    fR1 = afEB(0) * aafAbsC(2, 0) + afEB(1) * aafAbsC(2, 1) + afEB(2) * aafAbsC(2, 2) '
    fR01 = afEA(2) + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If

    ' Axes C0+t*B0
    fR = Abs(Dot(akB(0), kD)) '
    fR0 = afEA(0) * aafAbsC(0, 0) + afEA(1) * aafAbsC(1, 0) + afEA(2) * aafAbsC(2, 0) '
    fR01 = fR0 + afEB(0) '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If

    ' Axes C0+t*B1
    fR = Abs(Dot(akB(1), kD)) '
    fR0 = afEA(0) * aafAbsC(0, 1) + afEA(1) * aafAbsC(1, 1) + afEA(2) * aafAbsC(2, 1) '
    fR01 = fR0 + afEB(1) '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*B2
    fR = Abs(Dot(akB(2), kD)) '
    fR0 = afEA(0) * aafAbsC(0, 2) + afEA(1) * aafAbsC(1, 2) + afEA(2) * aafAbsC(2, 2) '
    fR01 = fR0 + afEB(2) '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Au moins une paire  d'axes est parralle
    If (ExistePaireParallele) Then
        TestIntersectionBoite = True '
        Exit Function
    End If
    
    ' Axes C0+t*A0xB0
    fR = Abs(afAD(2) * aafC(1, 0) - afAD(1) * aafC(2, 0)) '
    fR0 = afEA(1) * aafAbsC(2, 0) + afEA(2) * aafAbsC(1, 0) '
    fR1 = afEB(1) * aafAbsC(0, 2) + afEB(2) * aafAbsC(0, 1) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A0xB1
    fR = Abs(afAD(2) * aafC(1, 1) - afAD(1) * aafC(2, 1)) '
    fR0 = afEA(1) * aafAbsC(2, 1) + afEA(2) * aafAbsC(1, 1) '
    fR1 = afEB(0) * aafAbsC(0, 2) + afEB(2) * aafAbsC(0, 0) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A0xB2
    fR = Abs(afAD(2) * aafC(1, 2) - afAD(1) * aafC(2, 2)) '
    fR0 = afEA(1) * aafAbsC(2, 2) + afEA(2) * aafAbsC(1, 2) '
    fR1 = afEB(0) * aafAbsC(0, 1) + afEB(1) * aafAbsC(0, 0) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A1xB0
    fR = Abs(afAD(0) * aafC(2, 0) - afAD(2) * aafC(0, 0)) '
    fR0 = afEA(0) * aafAbsC(2, 0) + afEA(2) * aafAbsC(0, 0) '
    fR1 = afEB(1) * aafAbsC(1, 2) + afEB(2) * aafAbsC(1, 1) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A1xB1
    fR = Abs(afAD(0) * aafC(2, 1) - afAD(2) * aafC(0, 1)) '
    fR0 = afEA(0) * aafAbsC(2, 1) + afEA(2) * aafAbsC(0, 1) '
    fR1 = afEB(0) * aafAbsC(1, 2) + afEB(2) * aafAbsC(1, 0) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A1xB2
    fR = Abs(afAD(0) * aafC(2, 2) - afAD(2) * aafC(0, 2)) '
    fR0 = afEA(0) * aafAbsC(2, 2) + afEA(2) * aafAbsC(0, 2) '
    fR1 = afEB(0) * aafAbsC(1, 1) + afEB(1) * aafAbsC(1, 0) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A2xB0
    fR = Abs(afAD(1) * aafC(0, 0) - afAD(0) * aafC(1, 0)) '
    fR0 = afEA(0) * aafAbsC(1, 0) + afEA(1) * aafAbsC(0, 0) '
    fR1 = afEB(1) * aafAbsC(2, 2) + afEB(2) * aafAbsC(2, 1) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A2xB1
    fR = Abs(afAD(1) * aafC(0, 1) - afAD(0) * aafC(1, 1)) '
    fR0 = afEA(0) * aafAbsC(1, 1) + afEA(1) * aafAbsC(0, 1) '
    fR1 = afEB(0) * aafAbsC(2, 2) + afEB(2) * aafAbsC(2, 0) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    
    ' Axes C0+t*A2xB2
    fR = Abs(afAD(1) * aafC(0, 2) - afAD(0) * aafC(1, 2)) '
    fR0 = afEA(0) * aafAbsC(1, 2) + afEA(1) * aafAbsC(0, 2) '
    fR1 = afEB(0) * aafAbsC(2, 1) + afEB(1) * aafAbsC(2, 0) '
    fR01 = fR0 + fR1 '
    If (fR > fR01) Then
        TestIntersectionBoite = False '
        Exit Function
    End If
    

    TestIntersectionBoite = True
        
End Function
'----------------------------------------------------------------------------






